﻿using HeuristicLab.Algorithms.PushGP.Interpreter;
using HeuristicLab.Algorithms.PushGP.Stack;

namespace HeuristicLab.Algorithms.PushGP.Expressions
{
    /// <summary>
    /// Pushes a copy of an indexed item "deep" in the stack onto the top of the stack, without removing the deep item.
    /// The index is taken from the INTEGER stack, and the indexing is done after the index is removed.
    /// </summary>
    /// <typeparam name="T">Stacktype</typeparam>
    public abstract class YankDuplicateExpression<T> : Expression
    {
        public YankDuplicateExpression(string stringRepresentation) : base(stringRepresentation)
        { }

        public void Eval(IStack<T> stack, IStack<long> integerStack, int count = 1)
        {
            if (integerStack.Count == 0)
                return;

            var index = integerStack.Top;

            if (index > stack.Count - count ||
                index < 0 ||
                stack.Count < 2)
                return;

            integerStack.Pop();
            var item = stack.ElementAt((int)index);

            stack.Push(item);
        }
    }

    public class IntegerYankDuplicateExpression : YankDuplicateExpression<long>
    {
        public IntegerYankDuplicateExpression() : base("INTEGER.YANKDUP")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, interpreter.IntegerStack, 2);
        }
    }

    public class FloatYankDuplicateExpression : YankDuplicateExpression<double>
    {
        public FloatYankDuplicateExpression() : base("FLOAT.YANKDUP")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, interpreter.IntegerStack);
        }
    }

    public class BooleanYankDuplicateExpression : YankDuplicateExpression<bool>
    {
        public BooleanYankDuplicateExpression() : base("BOOLEAN.YANKDUP")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.BooleanStack, interpreter.IntegerStack);
        }
    }

    public class NameYankDuplicateExpression : YankDuplicateExpression<string>
    {
        public NameYankDuplicateExpression() : base("NAME.YANKDUP")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.NameStack, interpreter.IntegerStack);
        }
    }

    public class ExecYankDuplicateExpression : YankDuplicateExpression<Expression>
    {
        public ExecYankDuplicateExpression() : base("EXEC.YANKDUP")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.ExecStack, interpreter.IntegerStack);
        }
    }

    public class CodeYankDuplicateExpression : YankDuplicateExpression<Expression>
    {
        public CodeYankDuplicateExpression() : base("CODE.YANKDUP")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.CodeStack, interpreter.IntegerStack);
        }
    }
}
